@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2018 Snowflake Computing Inc. All right reserved.
+# Copyright (c) 2012-2019 Snowflake Computing Inc. All right reserved.
 #
 import collections
 import contextlib
@@ -9,35 +9,44 @@ import gzip
 import itertools
 import json
 import logging
-import platform
 import sys
 import time
+import traceback
 import uuid
-from io import StringIO, BytesIO
+from io import BytesIO
 from threading import Lock
 
 import OpenSSL.SSL
-from botocore.vendored import requests
-from botocore.vendored.requests.adapters import HTTPAdapter
-from botocore.vendored.requests.auth import AuthBase
-from botocore.vendored.requests.exceptions import (
+import requests
+from requests.adapters import HTTPAdapter
+from requests.auth import AuthBase
+from requests.exceptions import (
     ConnectionError, ConnectTimeout, ReadTimeout, SSLError)
-from botocore.vendored.requests.packages.urllib3.exceptions import (
+from requests.packages.urllib3.exceptions import (
     ProtocolError, ReadTimeoutError)
 
 from snowflake.connector.time_util import get_time_millis
 from . import ssl_wrap_socket
 from .compat import (
-    PY2,
     METHOD_NOT_ALLOWED, BAD_REQUEST, SERVICE_UNAVAILABLE, GATEWAY_TIMEOUT,
     FORBIDDEN, BAD_GATEWAY, REQUEST_TIMEOUT,
     UNAUTHORIZED, INTERNAL_SERVER_ERROR, OK, BadStatusLine)
-from .compat import (TO_UNICODE, urlencode, urlparse)
+from .compat import (TO_UNICODE, urlencode, urlparse, IncompleteRead)
 from .constants import (
     HTTP_HEADER_CONTENT_TYPE,
     HTTP_HEADER_ACCEPT,
     HTTP_HEADER_USER_AGENT,
-    HTTP_HEADER_SERVICE_NAME,
+    HTTP_HEADER_SERVICE_NAME
+)
+from .description import (
+    SNOWFLAKE_CONNECTOR_VERSION,
+    PYTHON_VERSION,
+    OPERATING_SYSTEM,
+    PLATFORM,
+    IMPLEMENTATION,
+    COMPILER,
+    CLIENT_NAME,
+    CLIENT_VERSION
 )
 from .errorcode import (ER_FAILED_TO_CONNECT_TO_DB, ER_CONNECTION_IS_CLOSED,
                         ER_FAILED_TO_REQUEST, ER_FAILED_TO_RENEW_SESSION)
@@ -46,20 +55,17 @@ from .errors import (Error, OperationalError, DatabaseError, ProgrammingError,
                      InterfaceError, InternalServerError, ForbiddenError,
                      BadGatewayError, BadRequest, MethodNotAllowed,
                      OtherHTTPRetryableError)
-from .gzip_decoder import decompress_raw_data
 from .sqlstate import (SQLSTATE_CONNECTION_NOT_EXISTS,
                        SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
-                       SQLSTATE_CONNECTION_REJECTED)
+                       SQLSTATE_CONNECTION_REJECTED,
+                       SQLSTATE_IO_ERROR)
+from .telemetry_oob import TelemetryService
 from .time_util import (
     DecorrelateJitterBackoff,
     DEFAULT_MASTER_VALIDITY_IN_SECONDS
 )
 from .tool.probe_connection import probe_connection
-from .util_text import split_rows_from_stream
-from .version import VERSION
 
-if PY2:
-    from pyasn1.error import PyAsn1Error
 
 logger = logging.getLogger(__name__)
 
@@ -102,19 +108,20 @@ REQUEST_ID = u'requestId'
 REQUEST_GUID = u'request_guid'
 SNOWFLAKE_HOST_SUFFIX = u'.snowflakecomputing.com'
 
-SNOWFLAKE_CONNECTOR_VERSION = u'.'.join(TO_UNICODE(v) for v in VERSION[0:3])
-PYTHON_VERSION = u'.'.join(TO_UNICODE(v) for v in sys.version_info[:3])
-OPERATING_SYSTEM = platform.system()
-PLATFORM = platform.platform()
-IMPLEMENTATION = platform.python_implementation()
-COMPILER = platform.python_compiler()
+SNOWFLAKE_CONNECTOR_VERSION = SNOWFLAKE_CONNECTOR_VERSION
+PYTHON_VERSION = PYTHON_VERSION
+OPERATING_SYSTEM = OPERATING_SYSTEM
+PLATFORM = PLATFORM
+IMPLEMENTATION = IMPLEMENTATION
+COMPILER = COMPILER
 
-CLIENT_NAME = u"PythonConnector"  # don't change!
-CLIENT_VERSION = u'.'.join([TO_UNICODE(v) for v in VERSION[:3]])
+CLIENT_NAME = CLIENT_NAME  # don't change!
+CLIENT_VERSION = CLIENT_VERSION
 PYTHON_CONNECTOR_USER_AGENT = \
-    u'{name}/{version}/{python_version}/{platform}'.format(
+    u'{name}/{version} ({platform}) {python_implementation}/{python_version}'.format(
         name=CLIENT_NAME,
         version=SNOWFLAKE_CONNECTOR_VERSION,
+        python_implementation=IMPLEMENTATION,
         python_version=PYTHON_VERSION,
         platform=PLATFORM)
 
@@ -184,24 +191,6 @@ class SnowflakeAuth(AuthBase):
         return r
 
 
-class ResultIterWithTimings(collections.Iterator):
-    DOWNLOAD = u"download"
-    PARSE = u"parse"
-
-    def __init__(self, it, timings):
-        self._it = it
-        self._timings = timings
-
-    def __next__(self):
-        return next(self._it)
-
-    def next(self):
-        return self.__next__()
-
-    def get_timings(self):
-        return self._timings
-
-
 class SnowflakeRestful(object):
     """
     Snowflake Restful class
@@ -220,13 +209,12 @@ class SnowflakeRestful(object):
         self._idle_sessions = collections.deque()
         self._active_sessions = set()
 
-        # insecure mode (disabled by default)
-        ssl_wrap_socket.FEATURE_INSECURE_MODE = \
-            self._connection and self._connection._insecure_mode
+        # OCSP mode (OCSPMode.FAIL_OPEN by default)
+        ssl_wrap_socket.FEATURE_OCSP_MODE = \
+            self._connection and self._connection._ocsp_mode()
         # cache file name (enabled by default)
         ssl_wrap_socket.FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME = \
             self._connection and self._connection._ocsp_response_cache_filename
-        #
 
         # This is to address the issue where requests hangs
         _ = 'dummy'.encode('idna').decode('utf-8')  # noqa
@@ -259,11 +247,6 @@ class SnowflakeRestful(object):
     def id_token(self, value):
         self._id_token = value
 
-    @property
-    def id_token_password(self):
-        return self._id_token_password if hasattr(
-            self, u'_id_token_password') else None
-
     def close(self):
         if hasattr(self, u'_token'):
             del self._token
@@ -271,8 +254,6 @@ class SnowflakeRestful(object):
             del self._master_token
         if hasattr(self, u'_id_token'):
             del self._id_token
-        if hasattr(self, u'_id_token_password'):
-            del self._id_token_password
         sessions = list(self._active_sessions)
         if sessions:
             logger.debug("Closing %s active sessions", len(sessions))
@@ -325,7 +306,7 @@ class SnowflakeRestful(object):
 
     def update_tokens(self, session_token, master_token,
                       master_validity_in_seconds=None,
-                      id_token=None, id_token_password=None):
+                      id_token=None):
         """
         Update session and master tokens and optionally temporary credential
         """
@@ -333,7 +314,6 @@ class SnowflakeRestful(object):
             self._token = session_token
             self._master_token = master_token
             self._id_token = id_token
-            self._id_token_password = id_token_password
             self._master_validity_in_seconds = master_validity_in_seconds
 
     def _renew_session(self):
@@ -396,8 +376,7 @@ class SnowflakeRestful(object):
                 ret[u'data'].get(u'masterToken'),
                 master_validity_in_seconds=ret[u'data'].get(
                     u'masterValidityInSeconds'),
-                id_token=self.id_token,
-                id_token_password=self.id_token_password)
+                id_token=self.id_token)
             logger.debug(u'updating session completed')
             ret[UPDATED_BY_ID_TOKEN] = request_type == REQUEST_TYPE_ISSUE
             return ret
@@ -447,7 +426,7 @@ class SnowflakeRestful(object):
             logger.error("Failed to heartbeat. code: %s, url: %s",
                          ret.get(u'code'), url)
 
-    def delete_session(self):
+    def delete_session(self, retry=False):
         """
         Deletes the session
         """
@@ -470,19 +449,31 @@ class SnowflakeRestful(object):
             headers[HTTP_HEADER_SERVICE_NAME] = self._connection.service_name
 
         body = {}
-        try:
-            ret = self._post_request(
-                url, headers, json.dumps(body),
-                token=self.token, timeout=5, no_retry=True)
-            if not ret or ret.get(u'success'):
-                return
-            err = ret.get(u'message')
-            if err is not None and ret.get(u'data'):
-                err += ret[u'data'].get(u'errorMessage', '')
-                # no exception is raised
-            logger.debug('error in deleting session. ignoring...: %s', err)
-        except Exception as e:
-            logger.debug('error in deleting session. ignoring...: %s', e)
+        retry_limit = 3 if retry else 1
+        num_retries = 0
+        should_retry = True
+        while should_retry and (num_retries < retry_limit):
+            try:
+                should_retry = False
+                ret = self._post_request(
+                    url, headers, json.dumps(body),
+                    token=self.token, timeout=5, no_retry=True)
+                if not ret:
+                    if retry:
+                        should_retry = True
+                    else:
+                        return
+                elif ret.get(u'success'):
+                    return
+                err = ret.get(u'message')
+                if err is not None and ret.get(u'data'):
+                    err += ret[u'data'].get(u'errorMessage', '')
+                    # no exception is raised
+                logger.debug('error in deleting session. ignoring...: %s', err)
+            except Exception as e:
+                logger.debug('error in deleting session. ignoring...: %s', e)
+            finally:
+                num_retries += 1
 
     def _get_request(self, url, headers, token=None,
                      timeout=None,
@@ -583,6 +574,7 @@ class SnowflakeRestful(object):
 
         class RetryCtx(object):
             def __init__(self, timeout, _include_retry_params=False):
+                self.total_timeout = timeout
                 self.timeout = timeout
                 self.cnt = 0
                 self.sleeping_time = 1
@@ -661,6 +653,19 @@ class SnowflakeRestful(object):
                 method, full_url, headers, data, conn)
             return {}
         except RetryRequest as e:
+            if retry_ctx.cnt == TelemetryService.get_instance().num_of_retry_to_trigger_telemetry:
+                _, _, stack_trace = sys.exc_info()
+                TelemetryService.get_instance().log_http_request_error(
+                    "HttpRequestRetry%dTimes" % retry_ctx.cnt,
+                    full_url,
+                    method,
+                    SQLSTATE_IO_ERROR,
+                    ER_FAILED_TO_REQUEST,
+                    retry_timeout=retry_ctx.total_timeout,
+                    retry_count=retry_ctx.cnt,
+                    exception=str(e),
+                    stack_trace=traceback.format_exc()
+                )
             if no_retry:
                 return {}
             cause = e.args[0]
@@ -668,6 +673,18 @@ class SnowflakeRestful(object):
                 retry_ctx.timeout -= int(time.time() - start_request_thread)
                 if retry_ctx.timeout <= 0:
                     logger.error(cause, exc_info=True)
+                    _, _, stack_trace = sys.exc_info()
+                    TelemetryService.get_instance().log_http_request_error(
+                        "HttpRequestRetryTimeout",
+                        full_url,
+                        method,
+                        SQLSTATE_IO_ERROR,
+                        ER_FAILED_TO_REQUEST,
+                        retry_timeout=retry_ctx.total_timeout,
+                        retry_count=retry_ctx.cnt,
+                        exception=str(e),
+                        stack_trace=traceback.format_exc()
+                    )
                     if isinstance(cause, Error):
                         Error.errorhandler_wrapper(conn, None, cause)
                     else:
@@ -700,7 +717,7 @@ class SnowflakeRestful(object):
         Error.errorhandler_wrapper(
             conn, None, OperationalError,
             {
-                u'msg': u'Failed to execute request: {0}'.format(
+                u'msg': u'Failed to execute request: {}'.format(
                     cause),
                 u'errno': ER_FAILED_TO_REQUEST,
             })
@@ -718,7 +735,7 @@ class SnowflakeRestful(object):
                     # masking the password
                     decoded_data['data']['PASSWORD'] = '********'
                     data = json.dumps(decoded_data)
-            except:
+            except Exception:
                 logger.info("data is not JSON")
         logger.error(
             u'Failed to get the response. Hanging? '
@@ -748,10 +765,8 @@ class SnowflakeRestful(object):
             catch_okta_unauthorized_error=False,
             is_raw_text=False,
             is_raw_binary=False,
-            is_raw_binary_iterator=True,
-            use_ijson=False,
-            socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT,
-            return_timing_metrics=False):
+            binary_data_handler=None,
+            socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT):
         if socket_timeout > DEFAULT_SOCKET_CONNECT_TIMEOUT:
             # socket timeout should not be more than the default.
             # A shorter timeout may be specified for login time, but
@@ -769,8 +784,7 @@ class SnowflakeRestful(object):
             else:
                 input_data = data
 
-            timing_metrics = {}
-            start_time = get_time_millis()
+            download_start_time = get_time_millis()
             # socket timeout is constant. You should be able to receive
             # the response within the time. If not, ConnectReadTimeout or
             # ReadTimeout is raised.
@@ -784,8 +798,7 @@ class SnowflakeRestful(object):
                 stream=is_raw_binary,
                 auth=SnowflakeAuth(token),
             )
-            timing_metrics[
-                ResultIterWithTimings.DOWNLOAD] = get_time_millis() - start_time
+            download_end_time = get_time_millis()
 
             try:
                 if raw_ret.status_code == OK:
@@ -793,23 +806,8 @@ class SnowflakeRestful(object):
                     if is_raw_text:
                         ret = raw_ret.text
                     elif is_raw_binary:
-                        start_time = get_time_millis()
-                        raw_data = decompress_raw_data(
-                            raw_ret.raw, add_bracket=True
-                        ).decode('utf-8', 'replace')
-                        if not is_raw_binary_iterator:
-                            ret = json.loads(raw_data)
-                        elif not use_ijson:
-                            ret = iter(json.loads(raw_data))
-                        else:
-                            ret = split_rows_from_stream(StringIO(raw_data))
-                        timing_metrics[
-                            ResultIterWithTimings.PARSE] = get_time_millis() - start_time
-
-                        # if timings requested, wrap the iterator so the timing
-                        # info is accessible
-                        if return_timing_metrics:
-                            ret = ResultIterWithTimings(ret, timing_metrics)
+                        ret = binary_data_handler.to_iterator(raw_ret.raw,
+                                                              download_end_time - download_start_time)
                     else:
                         ret = raw_ret.json()
                     return ret
@@ -838,6 +836,14 @@ class SnowflakeRestful(object):
                     )
                     return None  # required for tests
                 else:
+                    TelemetryService.get_instance().log_http_request_error(
+                        "HttpError%s" % str(raw_ret.status_code),
+                        full_url,
+                        method,
+                        SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
+                        ER_FAILED_TO_REQUEST,
+                        response=raw_ret
+                    )
                     Error.errorhandler_wrapper(
                         self._connection, None, InterfaceError,
                         {
@@ -854,11 +860,23 @@ class SnowflakeRestful(object):
                     return None  # required for tests
             finally:
                 raw_ret.close()  # ensure response is closed
+        except SSLError as se:
+            logger.debug("Hit non-retryable SSL error, %s", str(se))
+            TelemetryService.get_instance().log_http_request_error(
+                "CertificateException%s" % str(se),
+                full_url,
+                method,
+                SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
+                ER_FAILED_TO_REQUEST,
+                exception=se,
+                stack_trace=traceback.format_exc()
+            )
+
         except (ConnectTimeout,
                 ReadTimeout,
                 BadStatusLine,
                 ConnectionError,
-                SSLError,
+                IncompleteRead,
                 ProtocolError,  # from urllib3
                 ReadTimeoutError,  # from urllib3
                 OpenSSL.SSL.SysCallError,
@@ -873,12 +891,16 @@ class SnowflakeRestful(object):
                 exc_info=True)
             raise RetryRequest(err)
         except Exception as err:
-            if PY2 and isinstance(err, PyAsn1Error):
-                logger.debug(
-                    "Hit retryable client error. Retrying... "
-                    "Ignore the following error stack: %s", err,
-                    exc_info=True)
-                raise RetryRequest(err)
+            _, _, stack_trace = sys.exc_info()
+            TelemetryService.get_instance().log_http_request_error(
+                "HttpException%s" % str(err),
+                full_url,
+                method,
+                SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
+                ER_FAILED_TO_REQUEST,
+                exception=err,
+                stack_trace=traceback.format_exc()
+            )
             raise err
 
     def make_requests_session(self):
@@ -893,22 +915,30 @@ class SnowflakeRestful(object):
         """ Session caching context manager.  Note that the session is not
         closed until close() is called so each session may be used multiple
         times. """
-        try:
-            session = self._idle_sessions.pop()
-        except IndexError:
+        if self._connection.disable_request_pooling:
             session = self.make_requests_session()
-        self._active_sessions.add(session)
-        logger.debug("Active requests sessions: %s, idle: %s",
-                     len(self._active_sessions), len(self._idle_sessions))
-        try:
-            yield session
-        finally:
-            self._idle_sessions.appendleft(session)
             try:
-                self._active_sessions.remove(session)
-            except KeyError:
-                logger.debug(
-                    "session doesn't exist in the active session pool. "
-                    "Ignored...")
+                yield session
+            finally:
+                session.close()
+        else:
+            try:
+                session = self._idle_sessions.pop()
+            except IndexError:
+                session = self.make_requests_session()
+            self._active_sessions.add(session)
             logger.debug("Active requests sessions: %s, idle: %s",
                          len(self._active_sessions), len(self._idle_sessions))
+            try:
+                yield session
+            finally:
+                self._idle_sessions.appendleft(session)
+                try:
+                    self._active_sessions.remove(session)
+                except KeyError:
+                    logger.debug(
+                        "session doesn't exist in the active session pool. "
+                        "Ignored...")
+                logger.debug("Active requests sessions: %s, idle: %s",
+                             len(self._active_sessions),
+                             len(self._idle_sessions))
